/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Fabian Reister ( fabian dot reister at kit dot edu )
 * @date       2021
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#pragma once

#include <filesystem>
#include <map>
#include <vector>

#include <Eigen/Geometry>

#include <ArmarXCore/core/PackagePath.h>
#include <ArmarXCore/core/time/DateTime.h>

#include <RobotAPI/libraries/ArmarXObjects/ObjectID.h>
#include <RobotAPI/libraries/armem/core/Time.h>
#include <RobotAPI/libraries/armem_robot_state/aron/Proprioception.aron.generated.h>

namespace armarx::armem::robot_state
{

    namespace description
    {
        struct RobotDescription
        {
            // DateTime timestamp;

            std::string name;
            PackagePath xml{"", std::filesystem::path("")};
        };

        using RobotDescriptions = std::vector<RobotDescription>;

        std::ostream& operator<<(std::ostream& os, const RobotDescription& rhs);


    } // namespace description

    struct Twist
    {
        Eigen::Vector3f linear{Eigen::Vector3f::Zero()};
        Eigen::Vector3f angular{Eigen::Vector3f::Zero()};
    };

    struct PlatformState
    {
        Twist twist;
    };

    namespace proprioception
    {

        struct ForceTorque
        {
            Eigen::Vector3f force;
            Eigen::Vector3f gravityCompensatedForce;
            Eigen::Vector3f torque;
            Eigen::Vector3f gravityCompensatedTorque;
        };

        struct JointState
        {
            std::string name;
            float position;
        };

    } // namespace proprioception

    namespace exteroception
    {

        struct ToF
        {
            using DepthMatrixType = Eigen::MatrixXf; // FIXME uint16
            using DepthSigmaMatrixType = Eigen::MatrixXf; // FIXME uint8
            using TargetDetectedMatrixType = Eigen::MatrixXf; // FIXME uint8

            DepthMatrixType depth;
            DepthSigmaMatrixType sigma;
            TargetDetectedMatrixType targetDetected;

            std::string frame;
        };

    } // namespace exteroception

    struct RobotState
    {
        using JointMap = std::map<std::string, float>;
        using Pose = Eigen::Isometry3f;

        DateTime timestamp;

        Pose globalPose;
        JointMap jointMap;

        std::optional<armarx::armem::arondto::Proprioception> proprioception;
    };

    struct Robot
    {
        description::RobotDescription description;
        std::string instance;

        RobotState config;

        DateTime timestamp;

        std::string name() const;
    };

    using Robots = std::vector<Robot>;
    using RobotStates = std::vector<RobotState>;

    namespace localization
    {

        struct TransformHeader
        {
            std::string parentFrame;
            std::string frame;

            std::string agent;

            armem::Time timestamp;
        };

        struct Transform
        {
            TransformHeader header;

            Eigen::Isometry3f transform = Eigen::Isometry3f::Identity();

            EIGEN_MAKE_ALIGNED_OPERATOR_NEW
        };
    } // namespace localization


} // namespace armarx::armem::robot_state
